from tests.fixtures.artists.metadata import (
    Alias,
    AliasList,
    Area,
    Artist,
    BeginArea,
    EndArea,
    Gender,
    IpiList,
    IsniList,
    Iso31661CodeList,
    Iso31662CodeList,
    LifeSpan,
    Metadata,
)

__all__ = [
    "Alias",
    "AliasList",
    "Area",
    "Artist",
    "BeginArea",
    "EndArea",
    "Gender",
    "IpiList",
    "IsniList",
    "Iso31661CodeList",
    "Iso31662CodeList",
    "LifeSpan",
    "Metadata",
]
