{% include "imports.jinja2" %}
__all__ = [
{%- for source, items in imports|groupby("source") -%}
{%- for item in items %}
    "
    {%- if item.alias %}
        {{- item.alias|class_name -}}
    {% else %}
        {{- item.name|class_name -}}
    {% endif -%}
    ",
{%- endfor %}
{%- endfor %}
]
