from tests.fixtures.calculator.services import (
    Add,
    AddResponse,
    CalculatorSoapAdd,
    CalculatorSoapAddInput,
    CalculatorSoapAddOutput,
    CalculatorSoapDivide,
    CalculatorSoapDivideInput,
    CalculatorSoapDivideOutput,
    CalculatorSoapMultiply,
    CalculatorSoapMultiplyInput,
    CalculatorSoapMultiplyOutput,
    CalculatorSoapSubtract,
    CalculatorSoapSubtractInput,
    CalculatorSoapSubtractOutput,
    Divide,
    DivideResponse,
    Multiply,
    MultiplyResponse,
    Subtract,
    SubtractResponse,
)

__all__ = [
    "Add",
    "AddResponse",
    "CalculatorSoapAdd",
    "CalculatorSoapAddInput",
    "CalculatorSoapAddOutput",
    "CalculatorSoapDivide",
    "CalculatorSoapDivideInput",
    "CalculatorSoapDivideOutput",
    "CalculatorSoapMultiply",
    "CalculatorSoapMultiplyInput",
    "CalculatorSoapMultiplyOutput",
    "CalculatorSoapSubtract",
    "CalculatorSoapSubtractInput",
    "CalculatorSoapSubtractOutput",
    "Divide",
    "DivideResponse",
    "Multiply",
    "MultiplyResponse",
    "Subtract",
    "SubtractResponse",
]
