from tests.fixtures.series.series import (
    Links,
    Country,
    Externals,
    Image,
    Network,
    Previousepisode,
    Rating,
    Schedule,
    Self,
    Series,
)

__all__ = [
    "Links",
    "Country",
    "Externals",
    "Image",
    "Network",
    "Previousepisode",
    "Rating",
    "Schedule",
    "Self",
    "Series",
]
